use std::collections::HashMap;

use lazy_static::lazy_static;

use super::item::ItemId;

#[derive(Debug, thiserror::Error)]
#[error("{0} is not a dungeon super tile")]
pub struct NotADungeonSuperTile(u8);

#[derive(Clone, Copy, Eq, Hash, PartialEq)]
pub enum Dungeon {
    HyruleCastle,
    EasternPalace,
    DesertPalace,
    TowerOfHera,
    AgahnimsTower,
    PalaceOfDarkness,
    SwampPalace,
    SkullWoods,
    ThievesTown,
    IcePalace,
    MiseryMire,
    TurtleRock,
    GanonsTower,
}

pub fn get_dungeon_from_room(room_id: u8) -> Result<Dungeon, NotADungeonSuperTile> {
    for (dungeon, rooms) in DUNGEON_ROOMS.iter() {
        if let Ok(_) = rooms.binary_search(&room_id) {
            return Ok(*dungeon);
        }
    }

    Err(NotADungeonSuperTile(room_id))
}

const HYRULE_CASTLE_ROOMS: [u8; 21] = [
    1, 2, 17, 18, 33, 34, 50, 65, 66, 80, 81, 82, 96, 97, 98, 112, 113, 114, 128, 129, 130,
];

const EASTERN_PALACE_ROOMS: [u8; 13] = [
    137, 153, 168, 169, 170, 184, 185, 186, 200, 201, 216, 217, 218,
];

const DESERT_PALACE_ROOMS: [u8; 10] = [51, 67, 83, 99, 115, 116, 117, 131, 132, 133];

const TOWER_OF_HERA_ROOMS: [u8; 7] = [7, 23, 39, 49, 119, 135, 167];

const AGAHNIMS_TOWER_ROOMS: [u8; 7] = [32, 48, 64, 176, 192, 208, 224];

const PALACE_OF_DARKNESS_ROOMS: [u8; 14] = [9, 10, 11, 25, 26, 27, 42, 43, 58, 59, 74, 75, 90, 106];

const SWAMP_PALACE_ROOMS: [u8; 13] = [6, 22, 38, 40, 52, 53, 54, 55, 56, 70, 84, 102, 118];

const SKULL_WOODS_ROOMS: [u8; 9] = [41, 57, 73, 86, 87, 88, 89, 103, 104];

const THIEVES_TOWN_ROOMS: [u8; 12] = [68, 69, 100, 101, 171, 172, 187, 188, 203, 204, 219, 220];

const ICE_PALACE_ROOMS: [u8; 22] = [
    14, 30, 31, 46, 62, 63, 78, 79, 94, 95, 110, 126, 127, 142, 158, 159, 174, 175, 190, 191, 206,
    222,
];

const MISERY_MIRE_ROOMS: [u8; 18] = [
    144, 145, 146, 147, 151, 152, 160, 161, 162, 163, 177, 178, 179, 193, 194, 195, 209, 210,
];

const TURTLE_ROCK_ROOMS: [u8; 17] = [
    4, 19, 20, 21, 35, 36, 164, 180, 181, 182, 183, 196, 197, 198, 199, 213, 214,
];

const GANONS_TOWER_ROOMS: [u8; 26] = [
    12, 13, 28, 29, 61, 76, 77, 91, 92, 93, 107, 108, 109, 123, 124, 125, 139, 140, 141, 149, 150,
    155, 156, 157, 165, 166,
];

lazy_static! {
    static ref DUNGEON_ROOMS: HashMap<Dungeon, &'static [u8]> = {
        use Dungeon::*;

        let mut rooms = HashMap::new();

        rooms.insert(HyruleCastle, &HYRULE_CASTLE_ROOMS[..]);
        rooms.insert(EasternPalace, &EASTERN_PALACE_ROOMS[..]);
        rooms.insert(DesertPalace, &DESERT_PALACE_ROOMS[..]);
        rooms.insert(TowerOfHera, &TOWER_OF_HERA_ROOMS[..]);
        rooms.insert(AgahnimsTower, &AGAHNIMS_TOWER_ROOMS[..]);
        rooms.insert(PalaceOfDarkness, &PALACE_OF_DARKNESS_ROOMS[..]);
        rooms.insert(SwampPalace, &SWAMP_PALACE_ROOMS[..]);
        rooms.insert(SkullWoods, &SKULL_WOODS_ROOMS[..]);
        rooms.insert(ThievesTown, &THIEVES_TOWN_ROOMS[..]);
        rooms.insert(IcePalace, &ICE_PALACE_ROOMS[..]);
        rooms.insert(MiseryMire, &MISERY_MIRE_ROOMS[..]);
        rooms.insert(TurtleRock, &TURTLE_ROCK_ROOMS[..]);
        rooms.insert(GanonsTower, &GANONS_TOWER_ROOMS[..]);

        rooms
    };
    pub static ref DUNGEON_KEYS: HashMap<Dungeon, ItemId> = {
        use Dungeon::*;
        use ItemId::*;

        let mut keys = HashMap::new();

        keys.insert(HyruleCastle, HyruleKey);
        keys.insert(EasternPalace, EasternKey);
        keys.insert(DesertPalace, DesertKey);
        keys.insert(TowerOfHera, HeraKey);
        keys.insert(AgahnimsTower, AgaKey);
        keys.insert(PalaceOfDarkness, PodKey);
        keys.insert(SwampPalace, SwampKey);
        keys.insert(SkullWoods, SkullKey);
        keys.insert(ThievesTown, ThievesKey);
        keys.insert(IcePalace, IceKey);
        keys.insert(MiseryMire, MireKey);
        keys.insert(TurtleRock, TurtleKey);
        keys.insert(GanonsTower, GTKey);

        keys
    };
    pub static ref DUNGEON_BIG_KEYS: HashMap<Dungeon, ItemId> = {
        use Dungeon::*;
        use ItemId::*;

        let mut keys = HashMap::new();

        keys.insert(HyruleCastle, HyruleBigKey);
        keys.insert(EasternPalace, EasternBigKey);
        keys.insert(DesertPalace, DesertBigKey);
        keys.insert(TowerOfHera, HeraBigKey);
        keys.insert(AgahnimsTower, AgaBigKey);
        keys.insert(PalaceOfDarkness, PodBigKey);
        keys.insert(SwampPalace, SwampBigKey);
        keys.insert(SkullWoods, SkullBigKey);
        keys.insert(ThievesTown, ThievesBigKey);
        keys.insert(IcePalace, IceBigKey);
        keys.insert(MiseryMire, MireBigKey);
        keys.insert(TurtleRock, TurtleBigKey);
        keys.insert(GanonsTower, GTBigKey);

        keys
    };
}

#[derive(Clone, Copy, Eq, PartialEq, Hash)]
#[repr(u16)]
pub enum SuperTile {
    Ganon = 0x0,
    HyruleCastleNorthCorridor = 0x1,
    HyruleCastleSwitchRoom = 0x2,
    HoulihanRoom = 0x3,
    TurtleRockCrystalRollerRoom = 0x4,
    EmptyCloneRoom0x05 = 0x5,
    SwampPalaceArrghus = 0x6,
    TowerOfHeraMoldorm = 0x7,
    CaveHealingFairy = 0x8,
    PalaceOfDarknessMedusaChestRoom = 0x9,
    PalaceOfDarknessStalfossTrapRoom = 0xa,
    PalaceOfDarknessTurtleRoom = 0xb,
    GanonsTowerEntranceRoom = 0xc,
    GanonsTowerAgahnim2 = 0xd,
    IcePalaceEntranceRoom = 0xe,
    EmptyCloneRoom0x0F = 0xf,
    GanonEvacuationRoute = 0x10,
    HyruleCastleBombableStockRoom = 0x11,
    InsideSanctuary = 0x12,
    TurtleRockHokkuBokkuKeyRoom2 = 0x13,
    TurtleRockBigKeyRoom = 0x14,
    TurtleRockUselessTubes = 0x15,
    SwampPalaceSwimmingTreadmill = 0x16,
    TowerOfHeraMoldormFallRoom = 0x17,
    Cave0x18BigFairyDropEntrance = 0x18,
    PalaceOfDarknessDarkMaze = 0x19,
    PalaceOfDarknessBigChestRoom = 0x1a,
    PalaceOfDarknessMimicsMovingWallRoom = 0x1b,
    GanonsTowerIceArmos = 0x1c,
    GanonsTowerFinalHallway = 0x1d,
    IcePalaceBombFloorBariRoom = 0x1e,
    IcePalacePengatorBigKeyRoom = 0x1f,
    AgahnimsTowerAgahnim = 0x20,
    HyruleCastleKeyRatRoom = 0x21,
    HyruleCastleSewerTextTriggerRoom = 0x22,
    TurtleRockWestExitToBalcony = 0x23,
    TurtleRockDoubleHokkuBokkuBigChestRoom = 0x24,
    EmptyCloneRoom0x25 = 0x25,
    SwampPalaceStatueRoom = 0x26,
    TowerOfHeraBigChest = 0x27,
    SwampPalaceEntranceRoom = 0x28,
    SkullWoodsMothula = 0x29,
    PalaceOfDarknessBigHubRoom = 0x2a,
    PalaceOfDarknessMapChestFairyRoom = 0x2b,
    CaveHookshotCaveBackdoor = 0x2c,
    EmptyCloneRoom0x2D = 0x2d,
    IcePalaceCompassRoom = 0x2e,
    CaveKakarikoWellHP = 0x2f,
    AgahnimsTowerMaidenSacrificeChamber = 0x30,
    TowerOfHeraHardhatBeetleRoom = 0x31,
    HyruleCastleSewerKeyChestRoom = 0x32,
    DesertPalaceLanmolas = 0x33,
    SwampPalacePushBlockPuzzlePreBigKeyRoom = 0x34,
    SwampPalaceBigKeyBsRoom = 0x35,
    SwampPalaceBigChestRoom = 0x36,
    SwampPalaceMapChestWaterFillRoom = 0x37,
    SwampPalaceKeyPotRoom = 0x38,
    SkullWoodsGibdoKeyMothulaHoleRoom = 0x39,
    PalaceOfDarknessBombableFloorRoom = 0x3a,
    PalaceOfDarknessSpikeBlockConveyorRoom = 0x3b,
    CaveHookshotCave = 0x3c,
    GanonsTowerTorchRoom2 = 0x3d,
    IcePalaceStalfosKnightConveyorHellway = 0x3e,
    IcePalaceMapChestRoom = 0x3f,
    AgahnimsTowerFinalBridgeRoom = 0x40,
    HyruleCastleFirstDarkRoom = 0x41,
    HyruleCastle6RopesRoom = 0x42,
    DesertPalaceTorchPuzzleMovingWallRoom = 0x43,
    ThievesTownBigChestRoom = 0x44,
    ThievesTownJailCellsRoom = 0x45,
    SwampPalaceCompassChestRoom = 0x46,
    EmptyCloneRoom0x47 = 0x47,
    EmptyCloneRoom0x48 = 0x48,
    SkullWoodsGibdoTorchPuzzleRoom = 0x49,
    PalaceOfDarknessEntranceRoom = 0x4a,
    PalaceOfDarknessWarpsSouthMimicRoom = 0x4b,
    GanonsTowerMiniHelmasaurConveyorRoom = 0x4c,
    GanonsTowerMoldormRoom = 0x4d,
    IcePalaceBombJumpRoom = 0x4e,
    IcePalaceCloneRoomFairyRoom = 0x4f,
    HyruleCastleWestCorridor = 0x50,
    HyruleCastleThroneRoom = 0x51,
    HyruleCastleEastCorridor = 0x52,
    DesertPalacePopos2BeamosHellwayRoom = 0x53,
    SwampPalaceUpstairsPitsRoom = 0x54,
    CastleSecretEntranceUncleRoom = 0x55,
    SkullWoodsKeyPotTrapRoom = 0x56,
    SkullWoodsBigKeyRoom = 0x57,
    SkullWoodsBigChestRoom = 0x58,
    SkullWoodsFinalSectionEntranceRoom = 0x59,
    PalaceOfDarknessHelmasaurKing = 0x5a,
    GanonsTowerSpikePitRoom = 0x5b,
    GanonsTowerGanonBallZ = 0x5c,
    GanonsTowerGauntlet123 = 0x5d,
    IcePalaceLonelyFirebar = 0x5e,
    IcePalaceHiddenChestSpikeFloorRoom = 0x5f,
    HyruleCastleWestEntranceRoom = 0x60,
    HyruleCastleMainEntranceRoom = 0x61,
    HyruleCastleEastEntranceRoom = 0x62,
    DesertPalaceFinalSectionEntranceRoom = 0x63,
    ThievesTownWestAtticRoom = 0x64,
    ThievesTownEastAtticRoom = 0x65,
    SwampPalaceHiddenChestHiddenDoorRoom = 0x66,
    SkullWoodsCompassChestRoom = 0x67,
    SkullWoodsKeyChestTrapRoom = 0x68,
    EmptyCloneRoom0x69 = 0x69,
    PalaceOfDarknessRupeeRoom = 0x6a,
    GanonsTowerMimicsRoom = 0x6b,
    GanonsTowerLanmolasRoom = 0x6c,
    GanonsTowerGauntlet45 = 0x6d,
    IcePalacePengatorsRoom = 0x6e,
    EmptyCloneRoom0x6F = 0x6f,
    HyruleCastleSmallCorridorToJailCells = 0x70,
    HyruleCastleBoomerangChestRoom = 0x71,
    HyruleCastleMapChestRoom = 0x72,
    DesertPalaceBigChestRoom = 0x73,
    DesertPalaceMapChestRoom = 0x74,
    DesertPalaceBigKeyChestRoom = 0x75,
    SwampPalaceWaterDrainRoom = 0x76,
    TowerOfHeraEntranceRoom = 0x77,
    EmptyCloneRoom0x78 = 0x78,
    EmptyCloneRoom0x79 = 0x79,
    EmptyCloneRoom0x7A = 0x7a,
    GanonsTowerMisc = 0x7b,
    GanonsTowerEastSideCollapsingBridgeExplodingWallRoom = 0x7c,
    GanonsTowerWinderWarpMazeRoom = 0x7d,
    IcePalaceHiddenChestBombableFloorRoom = 0x7e,
    IcePalaceBigSpikeTrapsRoom = 0x7f,
    HyruleCastleJailCellRoom = 0x80,
    HyruleCastleNextToChasmRoom = 0x81,
    HyruleCastleBasementChasmRoom = 0x82,
    DesertPalaceWestEntranceRoom = 0x83,
    DesertPalaceMainEntranceRoom = 0x84,
    DesertPalaceEastEntranceRoom = 0x85,
    EmptyCloneRoom0x86 = 0x86,
    TowerOfHeraTileRoom = 0x87,
    EmptyCloneRoom0x88 = 0x88,
    EasternPalaceFairyRoom = 0x89,
    EmptyCloneRoom0x8A = 0x8a,
    GanonsTowerBlockPuzzleSpikeSkipMapChestRoom = 0x8b,
    GanonsTowerEastAndWestDownstairsBigChestRoom = 0x8c,
    GanonsTowerTileTorchPuzzleRoom = 0x8d,
    IcePalaceBlobsWithTetrisBarrier = 0x8e,
    EmptyCloneRoom0x8F = 0x8f,
    MiseryMireVitreous = 0x90,
    MiseryMireFinalSwitchRoom = 0x91,
    MiseryMireDarkBombWallSwitchesRoom = 0x92,
    MiseryMireDarkCaneFLoorSwitchPuzzleRoom = 0x93,
    EmptyCloneRoom0x94 = 0x94,
    GanonsTowerFinalCollapsingBridgeRoom = 0x95,
    GanonsTowerTorches1Room = 0x96,
    MiseryMireTorchPuzzleMovingWallRoom = 0x97,
    MiseryMireEntranceRoom = 0x98,
    EasternPalaceEyegoreKeyRoom = 0x99,
    EmptyCloneRoom0x9A = 0x9a,
    GanonsTowerManySpikesWarpMazeRoom = 0x9b,
    GanonsTowerInvisibleFloorMazeRoom = 0x9c,
    GanonsTowerCompassChestInvisibleFloorRoom = 0x9d,
    IcePalaceBigChestRoom = 0x9e,
    IcePalaceIceFloorPotsKeyAndSwitch = 0x9f,
    MiseryMirePreVitreousRoom = 0xa0,
    MiseryMireFishRoom = 0xa1,
    MiseryMireBridgeKeyChestRoom = 0xa2,
    MiseryMireEmptyLConnectingRoom = 0xa3,
    TurtleRockTrinexx = 0xa4,
    GanonsTowerWizzrobesRooms = 0xa5,
    GanonsTowerMoldormFallRoom = 0xa6,
    TowerOfHeraFairyRoom = 0xa7,
    EasternPalaceStalfosSpawnRoom = 0xa8,
    EasternPalaceBigChestRoom = 0xa9,
    EasternPalaceMapChestRoom = 0xaa,
    ThievesTownMovingSpikesKeyPotRoom = 0xab,
    ThievesTownBlindTheThief = 0xac,
    EmptyCloneRoom0xAD = 0xad,
    IcePalaceIceTRoom = 0xae,
    IcePalaceIceBridgeRoom = 0xaf,
    AgahnimsTowerCircleOfPots = 0xb0,
    MiseryMireHourglassRoom = 0xb1,
    MiseryMireSlugRoom = 0xb2,
    MiseryMireSpikeKeyChestRoom = 0xb3,
    TurtleRockPreTrinexxRoom = 0xb4,
    TurtleRockDarkMaze = 0xb5,
    TurtleRockChainChompsRoom = 0xb6,
    TurtleRockMapChestKeyChestRollerRoom = 0xb7,
    EasternPalaceBigKeyRoom = 0xb8,
    EasternPalaceLobbyCannonballsRoom = 0xb9,
    EasternPalaceDarkAntifairyKeyPotRoom = 0xba,
    ThievesTownHellway = 0xbb,
    ThievesTownConveyorToilet = 0xbc,
    EmptyCloneRoom0xBD = 0xbd,
    IcePalaceBlockPuzzleRoom = 0xbe,
    IcePalaceCloneRoomSwitchRoom = 0xbf,
    AgahnimsTowerDarkBridgeRoom = 0xc0,
    MiseryMireCompassChestTileRoom = 0xc1,
    MiseryMireBigHubRoom = 0xc2,
    MiseryMireBigChestRoom = 0xc3,
    TurtleRockFinalCrystalSwitchPuzzleRoom = 0xc4,
    TurtleRockLaserBridge = 0xc5,
    TurtleRockSomariaHub = 0xc6,
    TurtleRockTorchPuzzle = 0xc7,
    EasternPalaceArmosKnights = 0xc8,
    EasternPalaceEntranceRoom = 0xc9,
    UnknownRoom = 0xca,
    ThievesTownNorthWestEntranceRoom = 0xcb,
    ThievesTownNorthEastEntranceRoom = 0xcc,
    EmptyCloneRoom0xCD = 0xcd,
    IcePalaceHoleToKholdstareRoom = 0xce,
    EmptyCloneRoom0xCF = 0xcf,
    AgahnimsTowerDarkMaze = 0xd0,
    MiseryMireConveyorSlugBigKeyRoom = 0xd1,
    MiseryMireWizzrobesRoom = 0xd2,
    EmptyCloneRoom0xD3 = 0xd3,
    EmptyCloneRoom0xD4 = 0xd4,
    TurtleRockLaserKeyRoom = 0xd5,
    TurtleRockEntranceRoom = 0xd6,
    EmptyCloneRoom0xD7 = 0xd7,
    EasternPalacePreArmosKnightsRoom = 0xd8,
    EasternPalaceCannonballRoom = 0xd9,
    EasternPalacePotSwitchRoom = 0xda,
    ThievesTownSouthWestEntranceRoom = 0xdb,
    ThievesTownSouthEastEntranceRoom = 0xdc,
    EmptyCloneRoom0xDD = 0xdd,
    IcePalaceKholdstare = 0xde,
    CaveBackwardsDeathMountainTopFloor = 0xdf,
    AgahnimsTowerEntranceRoom = 0xe0,
    CaveLostWoodsHP = 0xe1,
    CaveLumberjacksTreeHP = 0xe2,
    CaveMagicBat = 0xe3,
    CaveLostOldManHouse = 0xe4,
    CaveLostOldManHouseBack = 0xe5,
    Cave0xE6 = 0xe6,
    Cave0xE7 = 0xe7,
    CaveSuperBunnyTop = 0xe8,
    EmptyCloneRoom0xE9 = 0xe9,
    CaveSpectacleRockHP = 0xea,
    CaveBumperCaveTop = 0xeb,
    EmptyCloneRoom0xEC = 0xec,
    Cave0xED = 0xed,
    CaveSpiralCave = 0xee,
    CaveCrystalSwitch5ChestsRoom = 0xef,
    CaveLostOldManStartingCaveBottom = 0xf0,
    CaveLostOldManStartingCaveTop = 0xf1,
    HouseOldWomanNextDoor = 0xf2,
    HouseOldWomanSahasrahlasWifeMaybe = 0xf3,
    HouseAngryBrothersWest = 0xf4,
    HouseAngryBrothersEast = 0xf5,
    EmptyCloneRoom0xF6 = 0xf6,
    EmptyCloneRoom0xF7 = 0xf7,
    CaveSuperBunnyBottom = 0xf8,
    CaveSpectacleRockExit = 0xf9,
    CaveSpectacleRockPrizeEntrance = 0xfa,
    CaveBumperCaveBottom = 0xfb,
    EmptyCloneRoom0xFC = 0xfc,
    Cave0xFD = 0xfd,
    CaveSpiralCaveExit = 0xfe,
    CaveParadoxCaveMiddleEntrance = 0xff,
    ShopInLostWoods0x100 = 0x100,
    ScaredLadyHouses = 0x101,
    SickKid = 0x102,
    InnBushHouse = 0x103,
    InsideLinksHouse = 0x104,
    ShabadooHouse = 0x105,
    ChestGameBombHouse = 0x106,
    LibraryBombFarmRoom = 0x107,
    ChickenHouse = 0x108,
    InsideWitchHut = 0x109,
    Aginah = 0x10a,
    Dam = 0x10b,
    MimicCave = 0x10c,
    CaveOutsideMiseryMire = 0x10d,
    CaveDarkHyliaLedgeHint = 0x10e,
    ShopDarkWorldShops = 0x10f,
    ShopFireShieldShop = 0x110,
    ArcherGame = 0x111,
    CaveShop = 0x112,
    KingsTomb = 0x113,
    WaterfallCaveMireHint = 0x114,
    BigFairy = 0x115,
    FatFairy = 0x116,
    SpikeCave = 0x117,
    ChestGamblingGame = 0x118,
    BlindsHouse = 0x119,
    Mutant = 0x11a,
    MirrorCaveGroveAndTomb = 0x11b,
    InsideBombShop = 0x11c,
    BlindsBasement = 0x11d,
    HypeCave = 0x11e,
    KakarikoShopLumberjackHouse = 0x11f,
    IceRodCave = 0x120,
    SmithHouse = 0x121,
    FortuneTellers = 0x122,
    MiniMoldormCave = 0x123,
    FiftyRupeeCaveBonkCave = 0x124,
    TwentyRupeeCaveDarkHyliaSpikeHintCave = 0x125,
    CheckerBoardCave = 0x126,
    HammerPegCave = 0x127,
    LostWoods = 0x8000,
    LumberjackHouse = 0x8002,
    DeathMountainWest = 0x8003,
    DeathMountainEast = 0x8005,
    DeathMountainTurtleWarp = 0x8007,
    EntranceDeathMountain = 0x800a,
    EntranceZorasDomain = 0x800f,
    PathFromLostWoodsToKakariko = 0x8010,
    KakarikoFortuneTeller = 0x8011,
    NorthWestWhirlpoolWarp = 0x8012,
    Sanctuary = 0x8013,
    Graveyard = 0x8014,
    River = 0x8015,
    WitchHut = 0x8016,
    EastOfWitchHut = 0x8017,
    Kakariko = 0x8018,
    CastleWoods = 0x801a,
    HyruleCastle = 0x801b,
    WitchGraveyardBridge = 0x801d,
    EasternPalace = 0x801e,
    Smithy = 0x8022,
    StonyFieldNorth = 0x8025,
    KakarikoMazeRace = 0x8028,
    KakarikoLibrary = 0x8029,
    HauntedGrove = 0x802a,
    BetweenGroveAndLinksHouse = 0x802b,
    LinksHouse = 0x802c,
    StonyFieldSouth = 0x802d,
    BetweenLakeHyliaAndEasternPalace = 0x802e,
    FluteSpot5 = 0x802f,
    Desert = 0x8030,
    SouthOfHauntedGrove = 0x8032,
    NorthwestSwamp = 0x8033,
    NortheastSwamp = 0x8034,
    LakeHylia = 0x8035,
    IceCave = 0x8037,
    DesertPass = 0x803a,
    SouthwestSwamp = 0x803b,
    SoutheastSwamp = 0x803c,
    SouthShore = 0x803f,
    ShullWoods = 0x8040,
    DarkLumberjackShop = 0x8042,
    DarkDeathMountainWest = 0x8043,
    DarkDeathMountainEast = 0x8045,
    TurtleRock = 0x8047,
    BumperCave = 0x804a,
    Catfish = 0x804f,
    PathFromSkullWoodsToVillage = 0x8050,
    OutcastFortuneTeller = 0x8051,
    NorthWestPond = 0x8052,
    DarkChapel = 0x8053,
    DarkGraveyard = 0x8054,
    DarkRiver = 0x8055,
    DarkWitchHut = 0x8056,
    EastOfDarkWitchHut = 0x8057,
    VillageOfOutcasts = 0x8058,
    FireShieldShopForest = 0x805a,
    Pyramid = 0x805b,
    BrokenBridge = 0x805d,
    PalaceOfDarkness = 0x805e,
    PurpleChestHouse = 0x8062,
    DarkStonyFieldNorth = 0x8065,
    DiggingGame = 0x8068,
    TrappedFrog = 0x8069,
    StumpysGrove = 0x806a,
    BetweenGroveAndBombShop = 0x806b,
    BombShop = 0x806c,
    HammerPegBridge = 0x806d,
    BetweenPODAndIceLake = 0x806e,
    DarkFluteSpot5 = 0x806f,
    MiseryMire = 0x8070,
    SouthOfStumpysGrove = 0x8072,
    DarkNorthwestSwamp = 0x8073,
    DarkNortheastSwamp = 0x8074,
    IceLake = 0x8075,
    DarkIceCave = 0x8077,
    BoxCanyonBetweenMireAndSwamp = 0x807a,
    DarkSouthwestSwamp = 0x807b,
    DarkSoutheastSwamp = 0x807c,
    DarkSouthShore = 0x807f,
    MasterSwordGladeAndUnderBridge = 0x8080,
    ZorasDomain = 0x8081,
}

use SuperTile::*;

lazy_static! {
    pub static ref ROOM_NAMES: HashMap<SuperTile, &'static str> = {

        let names = [
            (Ganon, "Ganon"),
            (HyruleCastleNorthCorridor, "Hyrule Castle (North Corridor)"),
            (HyruleCastleSwitchRoom, "Hyrule Castle (Switch Room)"),
            (HoulihanRoom, "Houlihan Room"),
            (TurtleRockCrystalRollerRoom, "Turtle Rock (Crystal Roller Room)"),
            (EmptyCloneRoom0x05, "Empty Clone Room"),
            (SwampPalaceArrghus, "Swamp Palace (Arrghus[Boss])"),
            (TowerOfHeraMoldorm, "Tower of Hera (Moldorm[Boss])"),
            (CaveHealingFairy, "Cave (Healing Fairy)"),
            (PalaceOfDarknessMedusaChestRoom, "Palace of Darkness (Medusa Chest Room)"),
            (PalaceOfDarknessStalfossTrapRoom, "Palace of Darkness (Stalfoss Trap Room)"),
            (PalaceOfDarknessTurtleRoom, "Palace of Darkness (Turtle Room)"),
            (GanonsTowerEntranceRoom, "Ganon's Tower (Entrance Room)"),
            (GanonsTowerAgahnim2, "Ganaon's Tower (Agahnim2[Boss])"),
            (IcePalaceEntranceRoom, "Ice Palace (Entrance Room)"),
            (EmptyCloneRoom0x0F, "Empty Clone Room"),
            (GanonEvacuationRoute, "Ganon Evacuation Route"),
            (HyruleCastleBombableStockRoom, "Hyrule Castle (Bombable Stock Room)"),
            (Sanctuary, "Sanctuary"),
            (TurtleRockHokkuBokkuKeyRoom2, "Turtle Rock (Hokku-Bokku Key Room 2)"),
            (TurtleRockBigKeyRoom, "Turtle Rock (Big Key Room)"),
            (TurtleRockUselessTubes, "Turtle Rock (Useless Tubes)"),
            (SwampPalaceSwimmingTreadmill, "Swamp Palace (Swimming Treadmill)"),
            (TowerOfHeraMoldormFallRoom, "Tower of Hera (Moldorm Fall Room)"),
            (Cave0x18BigFairyDropEntrance, "Big Fairy Drop Entrance Cave"),
            (PalaceOfDarknessDarkMaze, "Palace of Darkness (Dark Maze)"),
            (PalaceOfDarknessBigChestRoom, "Palace of Darkness (Big Chest Room)"),
            (PalaceOfDarknessMimicsMovingWallRoom, "Palace of Darkness (Mimics / Moving Wall Room"),
            (GanonsTowerIceArmos, "Ganon's Tower (Ice Armos)"),
            (GanonsTowerFinalHallway, "Ganon's Tower (Final Hallway)"),
            (IcePalaceBombFloorBariRoom, "Ice Palace (Bomb Floor / Bari Room)"),
            (IcePalacePengatorBigKeyRoom, "Ice Palace (Pengator / Big Key Room)"),
            (AgahnimsTowerAgahnim, "Agahnim's Tower (Agahnim[Boss])"),
            (HyruleCastleKeyRatRoom, "Hyrule Castle (Key Rat Room)"),
            (HyruleCastleSewerTextTriggerRoom, "Hyrule Castle (Sewer Text Trigger Room)"),
            (TurtleRockWestExitToBalcony, "Turtle Rock (West Exit to Balcony)"),
            (TurtleRockDoubleHokkuBokkuBigChestRoom, "Turtle Rock (Double Hokku-Bokku / Big Chest Room)"),
            (EmptyCloneRoom0x25, "Empty Clone Room"),
            (SwampPalaceStatueRoom, "Swamp Palace (Statue Room)"),
            (TowerOfHeraBigChest, "Tower of Hera (Big Chest)"),
            (SwampPalaceEntranceRoom, "Swamp Palace (Entrance Room)"),
            (SkullWoodsMothula, "Skull Woods (Mothula[Boss])"),
            (PalaceOfDarknessBigHubRoom, "Palace of Darkness (Big Hub Room)"),
            (PalaceOfDarknessMapChestFairyRoom, "Palace of Darkness (Map Chest / Fairy Room)"),
            (CaveHookshotCaveBackdoor, "Hookshot Cave Backdoor (Big Fairy)"),
            (EmptyCloneRoom0x2D, "Empty Clone Room"),
            (IcePalaceCompassRoom, "Ice Palace (Compass Room)"),
            (CaveKakarikoWellHP, "Kakariko Well HP"),
            (AgahnimsTowerMaidenSacrificeChamber, "Agahnim's Tower (Maiden Sacrifice Chamber)"),
            (TowerOfHeraHardhatBeetleRoom, "Tower of Hera (Hardhat Beetles Room)"),
            (HyruleCastleSewerKeyChestRoom, "Hyrule Castle (Sewer Key Chest Room)"),
            (DesertPalaceLanmolas, "Desert Palace (Lanmolas[Boss])"),
            (SwampPalacePushBlockPuzzlePreBigKeyRoom, "Swamp Palace (Push Block Puzzle / Pre-Big-Key Room"),
            (SwampPalaceBigKeyBsRoom, "Swamp Palace (Big Key / BS Room)"),
            (SwampPalaceBigChestRoom, "Swamp Palace (Big Chest Room)"),
            (SwampPalaceMapChestWaterFillRoom, "Swamp Palace (Map Chest / Water Fill Room)"),
            (SwampPalaceKeyPotRoom, "Swamp Palace (Key Pot Room)"),
            (SkullWoodsGibdoKeyMothulaHoleRoom, "Skull Woods (Gibdo Key / Mothula Hole Room)"),
            (PalaceOfDarknessBombableFloorRoom, "Palace of Darkness (Bombable Floor Room)"),
            (PalaceOfDarknessSpikeBlockConveyorRoom, "Palace of Darkness (Spike Block / Conveyor Room)"),
            (CaveHookshotCave, "Hookshot Cave"),
            (GanonsTowerTorchRoom2, "Ganon's Tower (Torch Room 2)"),
            (IcePalaceStalfosKnightConveyorHellway, "Ice Palace (Stalfos Knight / Conveyor Hellway)"),
            (IcePalaceMapChestRoom, "Ice Palace (Map Chest Room)"),
            (AgahnimsTowerFinalBridgeRoom, "Agahnim's Tower (Final Bridge Room)"),
            (HyruleCastleFirstDarkRoom, "Hyrule Castle (First Dark Room)"),
            (HyruleCastle6RopesRoom, "Hyrule Castle (6 Ropes Room)"),
            (DesertPalaceTorchPuzzleMovingWallRoom, "Desert Palace (Torch Puzzle / Moving Wall Room)"),
            (ThievesTownBigChestRoom, "Thieves' Town (Big Chest Room)"),
            (ThievesTownJailCellsRoom, "Thieves' Town (Jail Cells Room)"),
            (SwampPalaceCompassChestRoom, "Swamp Palace (Compass Chest Room)"),
            (EmptyCloneRoom0x47, "Empty Clone Room"),
            (EmptyCloneRoom0x48, "Empty Clone Room"),
            (SkullWoodsGibdoTorchPuzzleRoom, "Skull Woods (Gibdo Torch Puzzle Room)"),
            (PalaceOfDarknessEntranceRoom, "Palace of Darkness (Entrance Room)"),
            (PalaceOfDarknessWarpsSouthMimicRoom, "Palace of Darkness (Warps / South Mimic Room)"),
            (GanonsTowerMiniHelmasaurConveyorRoom, "Ganon's Tower (Mini-Helmasaur Conveyor Room)"),
            (GanonsTowerMoldormRoom, "Ganon's Tower (Moldorm Room)"),
            (IcePalaceBombJumpRoom, "Ice Palace (Bomb Jump Room)"),
            (IcePalaceCloneRoomFairyRoom, "Ice Palace Clone Room (Fairy Room)"),
            (HyruleCastleWestCorridor, "Hyrule Castle (West Corridor)"),
            (HyruleCastleThroneRoom, "Hyrule Castle (Throne Room)"),
            (HyruleCastleEastCorridor, "Hyrule Castle (East Corridor)"),
            (DesertPalacePopos2BeamosHellwayRoom, "Desert Palace (Popos 2 / Beamos Hellway Room)"),
            (SwampPalaceUpstairsPitsRoom, "Swamp Palace (Upstairs Pits Room)"),
            (CastleSecretEntranceUncleRoom, "Castle Secret Entrance / Uncle Room"),
            (SkullWoodsKeyPotTrapRoom, "Skull Woods (Key Pot / Trap Room)"),
            (SkullWoodsBigKeyRoom, "Skull Woods (Big Key Room)"),
            (SkullWoodsBigChestRoom, "Skull Woods (Big Chest Room)"),
            (SkullWoodsFinalSectionEntranceRoom, "Skull Woods (Final Section Entrance Room)"),
            (PalaceOfDarknessHelmasaurKing, "Palace of Darkness (Helmasaur King[Boss])"),
            (GanonsTowerSpikePitRoom, "Ganon's Tower (Spike Pit Room)"),
            (GanonsTowerGanonBallZ, "Ganon's Tower (Ganon-Ball Z)"),
            (GanonsTowerGauntlet123, "Ganon's Tower (Gauntlet 1/2/3)"),
            (IcePalaceLonelyFirebar, "Ice Palace (Lonely Firebar)"),
            (IcePalaceHiddenChestSpikeFloorRoom, "Ice Palace (Hidden Chest / Spike Floor Room)"),
            (HyruleCastleWestEntranceRoom, "Hyrule Castle (West Entrance Room)"),
            (HyruleCastleMainEntranceRoom, "Hyrule Castle (Main Entrance Room)"),
            (HyruleCastleEastEntranceRoom, "Hyrule Castle (East Entrance Room)"),
            (DesertPalaceFinalSectionEntranceRoom, "Desert Palace (Final Section Entrance Room)"),
            (ThievesTownWestAtticRoom, "Thieves' Town (West Attic Room)"),
            (ThievesTownEastAtticRoom, "Thieves' Town (East Attic Room)"),
            (SwampPalaceHiddenChestHiddenDoorRoom, "Swamp Palace (Hidden Chest / Hidden Door Room)"),
            (SkullWoodsCompassChestRoom, "Skull Woods (Compass Chest Room)"),
            (SkullWoodsKeyChestTrapRoom, "Skull Woods (Key Chest / Trap Room)"),
            (EmptyCloneRoom0x69, "Empty Clone Room"),
            (PalaceOfDarknessRupeeRoom, "Palace of Darkness (Rupee Room)"),
            (GanonsTowerMimicsRoom, "Ganon's Tower (Mimics Room)"),
            (GanonsTowerLanmolasRoom, "Ganon's Tower (Lanmolas Room"),
            (GanonsTowerGauntlet45, "Ganon's Tower (Gauntlet 4/5)"),
            (IcePalacePengatorsRoom, "Ice Palace (Pengators Room)"),
            (EmptyCloneRoom0x6F, "Empty Clone Room"),
            (HyruleCastleSmallCorridorToJailCells, "Hyrule Castle (Small Corridor to Jail Cells"),
            (HyruleCastleBoomerangChestRoom, "Hyrule Castle (Boomerang Chest Room)"),
            (HyruleCastleMapChestRoom, "Hyrule Castle (Map Chest Room)"),
            (DesertPalaceBigChestRoom, "Desert Palace (Big Chest Room)"),
            (DesertPalaceMapChestRoom, "Desert Palace (Map Chest Room)"),
            (DesertPalaceBigKeyChestRoom, "Desert Palace (Big Key Chest Room)"),
            (SwampPalaceWaterDrainRoom, "Swamp Palace (Water Drain Room)"),
            (TowerOfHeraEntranceRoom, "Tower of Hera (Entrance Room)"),
            (EmptyCloneRoom0x78, "Empty Clone Room"),
            (EmptyCloneRoom0x79, "Empty Clone Room"),
            (EmptyCloneRoom0x7A, "Empty Clone Room"),
            (GanonsTowerMisc, "Ganon's Tower (Sideways Conveyors / 4 Chest 4 Shooter / Square Pit"),
            (GanonsTowerEastSideCollapsingBridgeExplodingWallRoom, "Ganon's Tower (East Side Collapsing Bridge / Exploding Wall Room)"),
            (GanonsTowerWinderWarpMazeRoom, "Ganon's Tower (Winder / Warp Maze Room)"),
            (IcePalaceHiddenChestBombableFloorRoom, "Ice Palace (Hidden Chest / Bombable Floor Room)"),
            (IcePalaceBigSpikeTrapsRoom, "Ice Palace (Big Spike Traps Room)"),
            (HyruleCastleJailCellRoom, "Hyrule Castle (Jail Cell Room)"),
            (HyruleCastleNextToChasmRoom, "Hyrule Castle (Next to Chasm Room)"),
            (HyruleCastleBasementChasmRoom, "Hyrule Castle (Basement Chasm Room)"),
            (DesertPalaceWestEntranceRoom, "Desert Palace (West Entrance Room)"),
            (DesertPalaceMainEntranceRoom, "Desert Palace (Main Entrance Room)"),
            (DesertPalaceEastEntranceRoom, "Desert Palace (East Entrance Room)"),
            (EmptyCloneRoom0x86, "Empty Clone Room"),
            (TowerOfHeraTileRoom, "Tower of Hera (Tile Room)"),
            (EmptyCloneRoom0x88, "Empty Clone Room"),
            (EasternPalaceFairyRoom, "Eastern Palace (Fairy Room)"),
            (EmptyCloneRoom0x8A, "Empty Clone Room"),
            (GanonsTowerBlockPuzzleSpikeSkipMapChestRoom, "Ganon's Tower (Block Puzzle / Spike Skip / Map Chest Room)"),
            (GanonsTowerEastAndWestDownstairsBigChestRoom, "Ganon's Tower (East and West Downstairs / Big Chest Room)"),
            (GanonsTowerTileTorchPuzzleRoom, "Ganon's Tower (Tile / Torch Puzzle Room)"),
            (IcePalaceBlobsWithTetrisBarrier, "Ice Palace (Blobs with Tetris Barrier)"),
            (EmptyCloneRoom0x8F, "Empty Clone Room"),
            (MiseryMireVitreous, "Misery Mire (Vitreous[Boss])"),
            (MiseryMireFinalSwitchRoom, "Misery Mire (Final Switch Room)"),
            (MiseryMireDarkBombWallSwitchesRoom, "Misery Mire (Dark Bomb Wall / Switches Room)"),
            (MiseryMireDarkCaneFLoorSwitchPuzzleRoom, "Misery Mire (Dark Cane FLoor Switch Puzzle Room)"),
            (EmptyCloneRoom0x94, "Empty Clone Room"),
            (GanonsTowerFinalCollapsingBridgeRoom, "Ganon's Tower (Final Collapsing Bridge Room)"),
            (GanonsTowerTorches1Room, "Ganon's Tower (Torches 1 Room)"),
            (MiseryMireTorchPuzzleMovingWallRoom, "Misery Mire (Torch Puzzle / Moving Wall Room)"),
            (MiseryMireEntranceRoom, "Misery Mire (Entrance Room)"),
            (EasternPalaceEyegoreKeyRoom, "Eastern Palace (Eyegore Key Room)"),
            (EmptyCloneRoom0x9A, "Empty Clone Room"),
            (GanonsTowerManySpikesWarpMazeRoom, "Ganon's Tower (Many Spikes / Warp Maze Room)"),
            (GanonsTowerInvisibleFloorMazeRoom, "Ganon's Tower (Invisible Floor Maze Room)"),
            (GanonsTowerCompassChestInvisibleFloorRoom, "Ganon's Tower (Compass Chest / Invisible Floor Room)"),
            (IcePalaceBigChestRoom, "Ice Palace (Big Chest Room)"),
            (IcePalaceIceFloorPotsKeyAndSwitch, "Ice Palace (Room with ice floor, key, and 4 wall rats)"),
            (MiseryMirePreVitreousRoom, "Misery Mire (Pre-Vitreous Room)"),
            (MiseryMireFishRoom, "Misery Mire (Fish Room)"),
            (MiseryMireBridgeKeyChestRoom, "Misery Mire (Bridge Key Chest Room)"),
            (MiseryMireEmptyLConnectingRoom, "Misery Mire (Empty 'L' Connecting Room)"),
            (TurtleRockTrinexx, "Turtle Rock (Trinexx[Boss])"),
            (GanonsTowerWizzrobesRooms, "Ganon's Tower (Wizzrobes Rooms)"),
            (GanonsTowerMoldormFallRoom, "Ganon's Tower (Moldorm Fall Room)"),
            (TowerOfHeraFairyRoom, "Tower of Hera (Fairy Room)"),
            (EasternPalaceStalfosSpawnRoom, "Eastern Palace (Staflos Spawn Room)"),
            (EasternPalaceBigChestRoom, "Eastern Palace (Big Chest Room)"),
            (EasternPalaceMapChestRoom, "Eastern Palace (Map Chest Room)"),
            (ThievesTownMovingSpikesKeyPotRoom, "Thieves' Town (Moving Spikes / Key Pot Room)"),
            (ThievesTownBlindTheThief, "Thieves' Town (Blind the Thief[Boss])"),
            (EmptyCloneRoom0xAD, "Empty Clone Room"),
            (IcePalaceIceTRoom, "Ice Palace (Ice T Room)"),
            (IcePalaceIceBridgeRoom, "Ice Palace (Ice Bridge Room)"),
            (AgahnimsTowerCircleOfPots, "Agahnim's Tower (Circle of Pots)"),
            (MiseryMireHourglassRoom, "Misery Mire (Hourglass Room)"),
            (MiseryMireSlugRoom, "Misery Mire (Slug Room)"),
            (MiseryMireSpikeKeyChestRoom, "Misery Mire (Spike Key Chest Room)"),
            (TurtleRockPreTrinexxRoom, "Turtle Rock (Pre-Trinexx Room)"),
            (TurtleRockDarkMaze, "Turtle Rock (Dark Maze)"),
            (TurtleRockChainChompsRoom, "Turtle Rock (Chain Chomps Room)"),
            (TurtleRockMapChestKeyChestRollerRoom, "Turtle Rock (Map Chest / Key Chest / Roller Room)"),
            (EasternPalaceBigKeyRoom, "Eastern Palace (Big Key Room)"),
            (EasternPalaceLobbyCannonballsRoom, "Eastern Palace (Lobby Cannonballs Room)"),
            (EasternPalaceDarkAntifairyKeyPotRoom, "Eastern Palace (Dark Antifairy / Key Pot Room)"),
            (ThievesTownHellway, "Thieves' Town (Hellway)"),
            (ThievesTownConveyorToilet, "Thieves' Town (Conveyor Toilet)"),
            (EmptyCloneRoom0xBD, "Empty Clone Room"),
            (IcePalaceBlockPuzzleRoom, "Ice Palace (Block Puzzle Room)"),
            (IcePalaceCloneRoomSwitchRoom, "Ice Palace Clone Room (Switch Room)"),
            (AgahnimsTowerDarkBridgeRoom, "Agahnim's Tower (Dark Bridge Room)"),
            (MiseryMireCompassChestTileRoom, "Misery Mire (Compass Chest / Tile Room)"),
            (MiseryMireBigHubRoom, "Misery Mire (Big Hub Room)"),
            (MiseryMireBigChestRoom, "Misery Mire (Big Chest Room)"),
            (TurtleRockFinalCrystalSwitchPuzzleRoom, "Turtle Rock (Final Crystal Switch Puzzle Room)"),
            (TurtleRockLaserBridge, "Turtle Rock (Laser Bridge)"),
            (TurtleRockSomariaHub, "Turtle Rock (Somaria Hub Room)"),
            (TurtleRockTorchPuzzle, "Turtle Rock (Torch Puzzle)"),
            (EasternPalaceArmosKnights, "Eastern Palace (Armos Knights[Boss])"),
            (EasternPalaceEntranceRoom, "Eastern Palace (Entrance Room)"),
            (UnknownRoom, "Unused Room??"),
            (ThievesTownNorthWestEntranceRoom, "Thieves' Town (North West Entrance Room)"),
            (ThievesTownNorthEastEntranceRoom, "Thieves' Town (North East Entrance Room)"),
            (EmptyCloneRoom0xCD, "Empty Clone Room"),
            (IcePalaceHoleToKholdstareRoom, "Ice Palace (Hole to Kholdstare Room)"),
            (EmptyCloneRoom0xCF, "Empty Clone Room"),
            (AgahnimsTowerDarkMaze, "Agahnim's Tower (Dark Maze)"),
            (MiseryMireConveyorSlugBigKeyRoom, "Misery Mire (Conveyor Slug / Big Key Room)"),
            (MiseryMireWizzrobesRoom, "Misery Mire (Wizzrobes Room)"),
            (EmptyCloneRoom0xD3, "Empty Clone Room"),
            (EmptyCloneRoom0xD4, "Empty Clone Room"),
            (TurtleRockLaserKeyRoom, "Turtle Rock (Laser Key Room)"),
            (TurtleRockEntranceRoom, "Turtle Rock (Entrance Room)"),
            (EmptyCloneRoom0xD7, "Empty Clone Room"),
            (EasternPalacePreArmosKnightsRoom, "Eastern Palace (Pre-Armos-Knights Room)"),
            (EasternPalaceCannonballRoom, "Eastern Palace (Cannonball Room)"),
            (EasternPalacePotSwitchRoom, "Eastern Palace (Pot Switch Room)"),
            (ThievesTownSouthWestEntranceRoom, "Thieves' Town (South West Entrance Room)"),
            (ThievesTownSouthEastEntranceRoom, "Thieves Town (South East Entrance Room)"),
            (EmptyCloneRoom0xDD, "Empty Clone Room"),
            (IcePalaceKholdstare, "Ice Palace (Kholdstare[Boss])"),
            (CaveBackwardsDeathMountainTopFloor, "Death Mountain Exit (Top)"),
            (AgahnimsTowerEntranceRoom, "Agahnim's Tower (Entrance Room)"),
            (CaveLostWoodsHP, "Cave (Lost Woods HP)"),
            (CaveLumberjacksTreeHP, "Cave (Lumberjacks' Tree HP)"),
            (CaveMagicBat, "Cave (Magic Bat)"),
            (CaveLostOldManHouse, "Cave (Lost Old Man's House)"),
            (CaveLostOldManHouseBack, "Cave (Lost Old Man's House Back)"),
            (Cave0xE6, "Cave (Bunch of Keese)"),
            (Cave0xE7, "Cave (Bunch of Keese 2)"),
            (CaveSuperBunnyTop, "Cave (Super Bunny Top)"),
            (EmptyCloneRoom0xE9, "Empty Clone Room"),
            (CaveSpectacleRockHP, "Cave (Inside Spectacle Rock HP)"),
            (CaveBumperCaveTop, "Cave (Bumper Cave Top)"),
            (EmptyCloneRoom0xEC, "Empty Clone Room"),
            (Cave0xED, "Cave (??)"),
            (CaveSpiralCave, "Cave (Spiral Cave)"),
            (CaveCrystalSwitch5ChestsRoom, "Cave (Crystal Switch / 5 Chests Room)"),
            (CaveLostOldManStartingCaveBottom, "Cave (Lost Old Man Starting Cave Bottom)"),
            (CaveLostOldManStartingCaveTop, "Cave (Lost Old Man Starting Cave Top)"),
            (HouseOldWomanNextDoor, "House (Old Woman Next Door)"),
            (HouseOldWomanSahasrahlasWifeMaybe, "House (Old Woman / Sahasrahla's Wife?)"),
            (HouseAngryBrothersWest, "House (Angry Brothers West)"),
            (HouseAngryBrothersEast, "House (Angry Brothers East)"),
            (EmptyCloneRoom0xF6, "Empty Clone Room"),
            (EmptyCloneRoom0xF7, "Empty Clone Room"),
            (CaveSuperBunnyBottom, "Cave (Super Bunny Bottom)"),
            (CaveSpectacleRockExit, "Cave (Spectacle Rock Exit)"),
            (CaveSpectacleRockPrizeEntrance, "Cave (Spectacle Rock Prize Entrance)"),
            (CaveBumperCaveBottom, "Cave (Bumper Cave Bottom)"),
            (EmptyCloneRoom0xFC, "Empty Clone Room"),
            (Cave0xFD, "Cave (Death Mountain)"),
            (CaveSpiralCaveExit, "Cave (Spiral Cave Exit)"),
            (CaveParadoxCaveMiddleEntrance, "Cave (Paradox Cave 'Middle' Entrance)"),
            (ShopInLostWoods0x100, "Shop in Lost Woods"),
            (ScaredLadyHouses, "Scared Ladies' Houses"),
            (SickKid, "Sick Kid"),
            (InnBushHouse, "Inn / BushHouse"),
            (LinksHouse, "Link's House"),
            (ShabadooHouse, "Sahasrahla's House"),
            (ChestGameBombHouse, "Chest Game / Outcast Village Bomb House"),
            (LibraryBombFarmRoom, "Library / Bomb Farm Room"),
            (ChickenHouse, "Chicken House"),
            (WitchHut, "Witch Hut"),
            (Aginah, "Aginah's Cave"),
            (Dam, "Dam"),
            (MimicCave, "Mimic Cave"),
            (CaveOutsideMiseryMire, "Cave Outside Misery Mire"),
            (CaveDarkHyliaLedgeHint, "Cave (Dark Hylia Ledge Hint)"),
            (ShopDarkWorldShops, "Shop (Dark World)"),
            (ShopFireShieldShop, "Shop (Fire Shield Shop)"),
            (ArcherGame, "Archery Game"),
            (CaveShop, "Cave (Shop)"),
            (KingsTomb, "King's Tomb"),
            (WaterfallCaveMireHint, "Waterfall Cave / Cave (Mire Hint)"),
            (BigFairy, "Big Fairy"),
            (FatFairy, "Fat Fairy"),
            (SpikeCave, "Spike Cave"),
            (ChestGamblingGame, "Chest Gambling Game"),
            (BlindsHouse, "Blind's House"),
            (Mutant, "Mutant Hut"),
            (MirrorCaveGroveAndTomb, "Mirror Caves (South of Grove / Graveyard Ledge)"),
            (BombShop, "Bomb Shop"),
            (BlindsBasement, "Blind's Basement"),
            (HypeCave, "Hype Cave"),
            (KakarikoShopLumberjackHouse, "Kakariko Shop / Lumberjacks' House"),
            (IceRodCave, "Ice Rod Cave"),
            (SmithHouse, "Smiths' House"),
            (FortuneTellers, "Fortune Tellers"),
            (MiniMoldormCave, "Mini-Moldorm Cave"),
            (FiftyRupeeCaveBonkCave, "50 Rupee Cave / Bonk Cave"),
            (TwentyRupeeCaveDarkHyliaSpikeHintCave, "20 Rupee Cave / Dark Hylia Spike Hint Cave"),
            (CheckerBoardCave, "Checker Board Cave"),
            (HammerPegCave, "Hammer Peg Cave"),
        ];

        HashMap::from_iter(names.into_iter())
    };
}

pub const NEED_KILLABLE_ROOMS: [SuperTile; 39] = [
    EasternPalaceBigKeyRoom, // For anti-fairy circle around pot switch
    PalaceOfDarknessTurtleRoom,
    PalaceOfDarknessMimicsMovingWallRoom,
    PalaceOfDarknessWarpsSouthMimicRoom,
    TurtleRockCrystalRollerRoom, // Greed room has a kill-all section
    TurtleRockDoubleHokkuBokkuBigChestRoom,
    TurtleRockChainChompsRoom,
    SwampPalaceEntranceRoom,
    IcePalaceEntranceRoom, // Not normally required to backtrack, except in doors intensity 3
    IcePalaceCompassRoom,
    IcePalaceStalfosKnightConveyorHellway,
    IcePalacePengatorsRoom,
    TowerOfHeraHardhatBeetleRoom,
    TowerOfHeraTileRoom,
    ThievesTownBigChestRoom, // For trap room north of big chest
    ThievesTownJailCellsRoom, // For room with stairs and big lift block
    DesertPalacePopos2BeamosHellwayRoom,
    DesertPalaceBigKeyChestRoom,
    DesertPalaceEastEntranceRoom,
    GanonsTowerTorchRoom2,
    GanonsTowerGauntlet123,
    GanonsTowerMimicsRoom,
    GanonsTowerGauntlet45,
    GanonsTowerMisc,
    GanonsTowerWinderWarpMazeRoom,
    GanonsTowerTileTorchPuzzleRoom,
    GanonsTowerTorches1Room,
    GanonsTowerWizzrobesRooms,
    HyruleCastleBoomerangChestRoom,
    EasternPalaceStalfosSpawnRoom,
    EasternPalacePreArmosKnightsRoom,
    AgahnimsTowerCircleOfPots,
    AgahnimsTowerDarkBridgeRoom, // Not normally required to go backwards, except in basic/crossed doors, or doing glitch shenanigans.
    AgahnimsTowerEntranceRoom,
    MiseryMireSlugRoom,
    MiseryMireWizzrobesRoom,
    CaveCrystalSwitch5ChestsRoom,
    MimicCave,
    MiniMoldormCave,
];

// All the rooms that require special handling.

/// These rooms need to be locked to GFX ID 28.
pub const FREEZOR_ROOMS: [SuperTile; 5] = [
    IcePalaceEntranceRoom,
    IcePalaceHiddenChestBombableFloorRoom,
    IcePalaceBlobsWithTetrisBarrier,
    IcePalaceBigChestRoom,
    IcePalaceBlockPuzzleRoom,
];

/// These rooms need to be locked to GFX ID 17.
pub const WATER_ROOMS: [SuperTile; 33] = [
    GanonsTowerEntranceRoom,
    GanonsTowerAgahnim2,
    SwampPalaceSwimmingTreadmill,
    GanonsTowerIceArmos,
    GanonsTowerFinalHallway,
    SwampPalaceEntranceRoom,
    SwampPalacePushBlockPuzzlePreBigKeyRoom,
    SwampPalaceBigChestRoom,
    SwampPalaceKeyPotRoom,
    GanonsTowerTorchRoom2,
    SwampPalaceCompassChestRoom,
    GanonsTowerMiniHelmasaurConveyorRoom,
    GanonsTowerMoldormRoom,
    GanonsTowerSpikePitRoom,
    GanonsTowerGanonBallZ,
    GanonsTowerGauntlet123,
    SwampPalaceHiddenChestHiddenDoorRoom,
    GanonsTowerMimicsRoom,
    GanonsTowerLanmolasRoom,
    GanonsTowerGauntlet45,
    GanonsTowerMisc,
    GanonsTowerEastSideCollapsingBridgeExplodingWallRoom,
    GanonsTowerWinderWarpMazeRoom,
    GanonsTowerBlockPuzzleSpikeSkipMapChestRoom,
    GanonsTowerEastAndWestDownstairsBigChestRoom,
    GanonsTowerTileTorchPuzzleRoom,
    GanonsTowerFinalCollapsingBridgeRoom,
    GanonsTowerTorches1Room,
    GanonsTowerManySpikesWarpMazeRoom,
    GanonsTowerInvisibleFloorMazeRoom,
    GanonsTowerCompassChestInvisibleFloorRoom,
    GanonsTowerWizzrobesRooms,
    GanonsTowerMoldormFallRoom,
];

pub const SHADOW_ROOMS: [SuperTile; 2] = [
    IcePalaceStalfosKnightConveyorHellway,
    IcePalaceIceFloorPotsKeyAndSwitch,
];

pub const WALL_MASTER_ROOMS: [SuperTile; 5] = [
    SkullWoodsGibdoKeyMothulaHoleRoom,
    SkullWoodsGibdoTorchPuzzleRoom,
    SkullWoodsKeyPotTrapRoom,
    SkullWoodsKeyChestTrapRoom,
    GanonsTowerTileTorchPuzzleRoom,
];

pub const BUMPER_CRYSTAL_LASER_ROOMS: [SuperTile; 42] = [
    TurtleRockCrystalRollerRoom,
    PalaceOfDarknessTurtleRoom,
    TurtleRockHokkuBokkuKeyRoom2,
    TowerOfHeraMoldormFallRoom,
    PalaceOfDarknessMimicsMovingWallRoom,
    IcePalaceBombFloorBariRoom,
    PalaceOfDarknessBigHubRoom,
    PalaceOfDarknessMapChestFairyRoom,
    TowerOfHeraHardhatBeetleRoom,
    GanonsTowerTorchRoom2,
    AgahnimsTowerFinalBridgeRoom,
    ThievesTownBigChestRoom,
    PalaceOfDarknessEntranceRoom,
    GanonsTowerMiniHelmasaurConveyorRoom,
    SkullWoodsBigChestRoom,
    SkullWoodsFinalSectionEntranceRoom,
    GanonsTowerSpikePitRoom,
    SkullWoodsCompassChestRoom,
    SkullWoodsKeyChestTrapRoom,
    GanonsTowerMimicsRoom,
    TowerOfHeraEntranceRoom,
    IcePalaceHiddenChestBombableFloorRoom,
    TowerOfHeraTileRoom,
    GanonsTowerBlockPuzzleSpikeSkipMapChestRoom,
    MiseryMireFinalSwitchRoom,
    MiseryMireDarkBombWallSwitchesRoom,
    GanonsTowerTorches1Room,
    GanonsTowerManySpikesWarpMazeRoom,
    GanonsTowerCompassChestInvisibleFloorRoom,
    MiseryMireFishRoom,
    GanonsTowerWizzrobesRooms,
    ThievesTownMovingSpikesKeyPotRoom,
    TurtleRockChainChompsRoom,
    IcePalaceCloneRoomSwitchRoom,
    MiseryMireCompassChestTileRoom,
    MiseryMireBigChestRoom,
    TurtleRockFinalCrystalSwitchPuzzleRoom,
    TurtleRockLaserBridge,
    TurtleRockLaserKeyRoom,
    TurtleRockEntranceRoom,
    CaveBumperCaveTop,
    CaveCrystalSwitch5ChestsRoom,
];

pub const PULL_SWITCH_ROOMS: [SuperTile; 4] = [
    HyruleCastleSwitchRoom,
    SkullWoodsBigChestRoom,
    ThievesTownWestAtticRoom,
    Dam,
];

pub const TONGUE_ROOMS: [SuperTile; 7] = [
    TurtleRockCrystalRollerRoom,
    TurtleRockWestExitToBalcony,
    SwampPalaceBigKeyBsRoom,
    SwampPalaceMapChestWaterFillRoom,
    IcePalaceMapChestRoom,
    SwampPalaceWaterDrainRoom,
    IcePalaceHoleToKholdstareRoom,
];

pub const NO_STATUE_ROOMS: [SuperTile; 12] = [
    SwampPalaceSwimmingTreadmill,
    SwampPalaceStatueRoom,
    SwampPalaceEntranceRoom,
    PalaceOfDarknessMapChestFairyRoom,
    SwampPalacePushBlockPuzzlePreBigKeyRoom,
    SwampPalaceBigChestRoom,
    SwampPalaceCompassChestRoom,
    SkullWoodsBigKeyRoom,
    SwampPalaceWaterDrainRoom,
    MiseryMireHourglassRoom,
    MiseryMireBigHubRoom,
    AgahnimsTowerDarkMaze,
];

pub const MOVING_CANNONS_ROOMS: [SuperTile; 2] = [
    GanonsTowerGanonBallZ,
    DesertPalaceBigKeyChestRoom,
];

pub const CANNON_ROOMS: [SuperTile; 3] = [
    EasternPalaceLobbyCannonballsRoom,
    EasternPalaceCannonballRoom,
    MimicCave, // ???
];

/// These rooms have no enemies to randomize, or are special bosses that can't be switched out.
pub const DONT_RANDOMIZE_ROOMS: [SuperTile; 8] = [
    Ganon,
    HyruleCastleNorthCorridor,
    HoulihanRoom,
    GanonsTowerAgahnim2,
    TurtleRockBigKeyRoom,
    AgahnimsTowerAgahnim,
    AgahnimsTowerMaidenSacrificeChamber,
    IcePalaceBigSpikeTrapsRoom, // Anything else would be too easy?
];

pub const NO_SPECIAL_ENEMIES_IN_STANDARD_MODE_ROOMS: [SuperTile; 21] = [
    HyruleCastleNorthCorridor,
    HyruleCastleSwitchRoom,
    HyruleCastleBombableStockRoom,
    HyruleCastleKeyRatRoom,
    HyruleCastleSewerTextTriggerRoom,
    HyruleCastleSewerKeyChestRoom,
    HyruleCastleFirstDarkRoom,
    HyruleCastle6RopesRoom,
    HyruleCastleWestCorridor,
    HyruleCastleThroneRoom,
    HyruleCastleEastCorridor,
    CastleSecretEntranceUncleRoom,
    HyruleCastleWestEntranceRoom,
    HyruleCastleMainEntranceRoom,
    HyruleCastleEastEntranceRoom,
    HyruleCastleSmallCorridorToJailCells,
    HyruleCastleBoomerangChestRoom,
    HyruleCastleMapChestRoom,
    HyruleCastleJailCellRoom,
    HyruleCastleNextToChasmRoom,
    HyruleCastleBasementChasmRoom,
];

pub const BOSS_ROOMS: [SuperTile; 16] = [
    Ganon,
    SwampPalaceArrghus,
    TowerOfHeraMoldorm,
    GanonsTowerAgahnim2,
    GanonsTowerIceArmos,
    AgahnimsTowerAgahnim,
    SkullWoodsMothula,
    DesertPalaceLanmolas,
    GanonsTowerMoldormRoom,
    PalaceOfDarknessHelmasaurKing,
    GanonsTowerLanmolasRoom,
    MiseryMireVitreous,
    TurtleRockTrinexx,
    ThievesTownBlindTheThief,
    EasternPalaceArmosKnights,
    IcePalaceKholdstare,
];

pub const DONT_USE_IMMOVABLE_ENEMIES: [SuperTile; 57] = [
    PalaceOfDarknessTurtleRoom, // for single terrorpin in L shaped section
    SwampPalaceSwimmingTreadmill,
    PalaceOfDarknessDarkMaze, // top placement will probably block maze
    IcePalaceBombFloorBariRoom, // TODO: test
    SwampPalaceStatueRoom,
    TowerOfHeraBigChest, // TODO: test, top left placement before stairs up
    SwampPalaceBigChestRoom, // bottom left waterbug
    IcePalaceMapChestRoom, // spikes block stuff
    HyruleCastle6RopesRoom, // only if two of them stack, but why chance it?
    AgahnimsTowerFinalBridgeRoom, // spikes are bad here
    SwampPalaceCompassChestRoom,
    SkullWoodsGibdoTorchPuzzleRoom,
    PalaceOfDarknessWarpsSouthMimicRoom,
    IcePalaceBombJumpRoom, // for the zols in the small next room
    CastleSecretEntranceUncleRoom, // TODO: test
    SkullWoodsBigKeyRoom,
    IcePalaceHiddenChestSpikeFloorRoom, // TODO: would this cause problems in OHKO since you can't hookshot across if middle mob is beamos, etc?
    ThievesTownEastAtticRoom, // only if both bottom rats
    PalaceOfDarknessRupeeRoom, // only if two terrorpins next to each other (but can happen twice)
    DesertPalaceMapChestRoom, // only if both antlions in hallway to the south of chest
    SwampPalaceWaterDrainRoom, // would need 3 mobs to be impassible to possibly softlock...
    GanonsTowerWinderWarpMazeRoom, // very low chance
    IcePalaceBigSpikeTrapsRoom, // TODO: what happens to beamos over a pit?
    DesertPalaceWestEntranceRoom, // TODO: test
    DesertPalaceMainEntranceRoom, // TODO: test
    DesertPalaceEastEntranceRoom, // TODO: test
    GanonsTowerEastAndWestDownstairsBigChestRoom, // TODO: test, probably safe?
    GanonsTowerTileTorchPuzzleRoom, // TODO: test
    MiseryMireDarkBombWallSwitchesRoom, // TODO: test
    GanonsTowerFinalCollapsingBridgeRoom, // TODO: test, probably safe because of conveyor belts
    MiseryMireEntranceRoom,
    GanonsTowerManySpikesWarpMazeRoom, // TODO: test, middle spike covers warp, are we randomizing those?
    GanonsTowerInvisibleFloorMazeRoom, // TODO: test
    GanonsTowerCompassChestInvisibleFloorRoom, // TODO: test
    IcePalaceBigChestRoom, // Big spikes will block
    MiseryMirePreVitreousRoom, // TODO: test
    EasternPalaceMapChestRoom, // TODO: test
    IcePalaceIceBridgeRoom,
    MiseryMireSpikeKeyChestRoom, // TODO: test lower stalfos blocking door
    EasternPalaceDarkAntifairyKeyPotRoom, // TODO: test lower stalfos blocking door
    ThievesTownHellway, // TODO: test, but should be ok
    ThievesTownConveyorToilet, // TODO: test
    TurtleRockSomariaHub, // technically a door is blocked off, but who would ever go there? (Door rando would)
    ThievesTownNorthWestEntranceRoom,
    IcePalaceHoleToKholdstareRoom, // spikes block stuff
    AgahnimsTowerDarkMaze, // TODO: test
    MiseryMireWizzrobesRoom,
    TurtleRockLaserKeyRoom,
    EasternPalacePreArmosKnightsRoom,
    ThievesTownSouthEastEntranceRoom, // TODO: test
    CaveBackwardsDeathMountainTopFloor,
    CaveLostOldManHouse,
    Cave0xE7,
    CaveSpiralCave,
    CaveSpectacleRockExit, // TODO: test, probably can get past
    Cave0xFD,
    MimicCave,
];

pub const DONT_USE_FLYING_ENEMIES: [SuperTile; 2] = [
    MiseryMireWizzrobesRoom,
    MimicCave,
];
